% run_experiments.m: -> CreateModel() loads FBA models and defines metabolite exchange.
%
%
% Input:    (1)'FBAmodel_path' folder
%           (2)dictionary file
%           (3)species names
%           (4)parameters
%
% Output:   empty structures of coculture and models{k}, not yet initialized


function [coculture , models] = CreateModel(FBAmodel_path, Dictionary ,species, params)

%#########################################################################
% (1) load models

    for z=1:length(species)
        file_name = species{z};
        load( fullfile(FBAmodel_path, file_name) );
        species_name = strrep(file_name,'.mat','');
        % Create a local copy

        eval(['newf = fields(' species_name ');']);
        for i = 1:length(newf)
            models(z).(newf{i}) = eval([species_name '.(newf{i})']);
        end

        coculture.species{z} = species_name;
        if (~isfield(models(z),'description')) || isempty(models(z).description)
            models(z).description = species_name;
        end
    end
    % structure need to be exactly the same so had to be added in separate
    % loop without loading next models
    for z=1:length(species)
        models(z).species = coculture.species{z};
    end


%#########################################################################
% (2) DefineExchange
    mets_universe={};
    for k=1:length(coculture.species)
       % Find exchange reactions from S matrix
        models(k).isEX = zeros(length(models(k).rxns),1);

        tmp = sum( models(k).S ~=0 );      % number of mets involved in each rxn
        models(k).isEX(tmp ==1) = 1; % EX_v only has one met involved

        %exclude objective
        obj = find(models(k).c);
        models(k).isEX(obj)=0;
    % % Patch for SEED reactions that occurs for free(unknown components):  % %
    % % This issue has not been solved generically.    Hsuanchao 2012/8/20  % %
    % % rxn13782: Nothing => Protein biosynthesis
    % % rxn13783: Nothing => DNA replication
    % % rxn13784: Nothing => RNA transcription
    % % These rxns has coeff=+1 in S matrix while normal EXrxns, or sinks has -1
    % % bio00107: Others + Protein biosynthesis + DNA replication + RNA transcription => Biomass + Others

		SEEDexception_rxns = {'rxn13782','rxn13783','rxn13784'};
		SEEDexception_rxnNames = {'Protein biosynthesis','DNA replication','RNA transcription'};
		for i=1:length(SEEDexception_rxns)
			clear a;
			a = strmatch(SEEDexception_rxns{i} , models(k).rxns, 'exact');
			b = strmatch(SEEDexception_rxnNames{i} , models(k).rxnNames, 'exact');
			if(~isempty(a) && ~isempty(b) && (a==b))
				models(k).isEX(a) = 0;
			end
		end

        a = find(models(k).isEX);
        %Exclude Biomass testing sink, for bug on Toy model 9/24/2012
        a = setdiff(a, find(models(k).c) );
        for i=1:length(a)
            r = a(i);
            rxns_r = char(models(k).rxns( r ));
            c = min(3,length(rxns_r));
            if (strcmp(rxns_r(1:c) ,'EX_') == 0)

                %Known model sink/unidirectional production reactions
                sink_iAF1260={'DM_4HBA','DM_5DRIB','DM_AACALD','DM_HMFURN','DM_OXAM'};
                sink_iJO1366={'DM_4CRSOL','DM_5DRIB','DM_AACALD','DM_AMOB','DM_MTHTHF','DM_OXAM'};
                sink_iAH991 = {'DM_4HBA','DM_5DRIB','DM_AMOB','sink_chols','sink_s'};
                sink_iEre400 = {'co2Out','butOut','r111','r113','H2Out'};

                SEEDexception_rxns = {'rxn13782_c0','rxn13783_c0','rxn13784_c0'};
                AGORAexception_rxns = {'DM_dhptd(c)','dreplication','pbiosynthesis','rtranscription','sink_PGPm1[c]','sink_dmbzid',...
                    'DM_4HBA','DM_HQN','DM_clpn140(c)','DM_clpn160(c)','DM_clpn180(c)','DM_clpnai15(c)','DM_clpnai17(c)',...
                    'DM_clpni15(c)','DM_clpni16(c)','DM_clpni17(c)','DM_5MTR','DM_dad_5','sink_gthrd(c)','DM_2HYMEPH','DM_teich_45_BS(c)',...
                    'DM_hcys_L[c]','DM_btn','DM_kdo2lipid4L(c)','DM_adocbl(c)','DM_pheme(c)'};
                s1 = union(sink_iAF1260,sink_iJO1366);
                s2 = union(sink_iAH991, sink_iEre400);

                exception = union(s1,s2);
                exception = union(exception,SEEDexception_rxns);
                exception = union(exception, AGORAexception_rxns);
                %An uptake reaction that we don't recognize
                L = models(k).lb(r);
                U = models(k).ub(r);
                rev = L*U;
                rxnName = models(k).rxnNames(r);

                %Unidirectional (sink or secretion) reactions are not
                %consider as EXrxns
                if ismember(rxns_r,exception) && (rev <= 0)
                    x = models(k).isEX(r);
                    models(k).isEX(r) = 0;
                else%if (rev >= 0)
                    error(['An Exchange rxn not named as EX_xxx, check this model "' models(k).species '"' ':   ' models(k).rxnNames(r) ]);
                end
            end
        end





    % % Patch for SEED reactions that occurs for free(unknown components):  % %

        EX_rxnid = find( models(k).isEX==1 );

       % figure out which met is secreted by these rxns
       %   find which met is secreted by flux i
       %   this is likely the place the user medium_mets mapping files gets in
       clear  metsID;
	   metsID = zeros(1,length(  EX_rxnid  ));
       for i=1:length(  EX_rxnid  )
            metsID(i) = find(models(k).S(:   ,EX_rxnid(i)    ));

              str = char(models(k).mets(metsID(i))  );
              % determine if this metabolite is a sink metabolite
              a = 0;
              if(length(str)>=3)
                  a = str(end-2:end);
              end
              EXmets_is_sink(k,i) = 1 - strcmp('[e]', a  ) ;
       end
       EXmets = models(k).mets(metsID);
       EXmetNames = models(k).metNames(metsID);
       % Klitgord data patch
       % E.coli iAF1260 has sink metabolites 4hba[c] ==> nothing
       % We can (1)Deletes these sink metabolites or exclude from medium
       %        EX_rxnid(  find(EXmets_is_sink(k,:)) ) =[];
       %        metsID(  find(EXmets_is_sink(k,:)) ) =[];
       %        EXmets(  find(EXmets_is_sink(k,:)) ) =[];
       %        (2)Rename them (e.g. 4hba_sink[c]) to track them
       %        id = find(EXmets_is_sink(k,:));
       %        for i=1:length(id)
       %            j = id(i);
       %            str = char(EXmets(j));
       %            str = {[str(1:end-3) '_SINKED'     str(end-2:end) ]};
       %            EXmets(j) = cell(str);
       %        end
       % NOW I prefer leave it as user's job, judge it afterwards

       mets_universe = [mets_universe ;EXmets];
       models(k).EXmets = EXmets;  % This is the name of mets secreted by the EX rxns
       models(k).EXmetNames = EXmetNames;
       models(k).EXmetsID = metsID;
       models(k).EXrxnsID = EX_rxnid;
       models(k).EXrxns = models(k).rxns(EX_rxnid);
    end



    % ----- Dictionary for metabolite mapping in each FBA model and in medium -----
    MetsMappingFile = fullfile(params.input_dir, Dictionary); % This file has to follow the sequential order of FBA models
    % MetsMappingFile:
    %                 Organism
    %           Eco     Bsu  Medium
    %        -----------------------
    %  mets  1|  a      A      h2o
    % vector 2|  atp    atp    atp
    %        3|  c      -      attn
    %        4|  -      nad    nad
    %        5|  ala_L  ala-L  ala-L
    %  coculture.mets is the last column := {h2o , atp , attn, nad , ala-L}
    %  mets_dictionary is the first two columns:= { a A ; atp atp ; c NaN ; NaN  nad ; ala_L ala-L}
    %  coculture.mets_alias is the one different than medium column := {a  A  c  ala_L};
    %  coculture.mets_alias_hashid is the index of medium id(row number) := [1  1  3    5  ];
    %
    %  later in the code, compartment infomation (e) or [e] will be rename into[env]
    %

	fid = fopen(MetsMappingFile);
	N = length(coculture.species);
	tline = fgetl(fid);
	species_name_medium = textscan(tline,'%s','delimiter',',');
	species_name_medium = species_name_medium{1}';
	mets_dictionary=[];
	m=1;
	while ~feof(fid)
		tline = fgetl(fid);
		if ~isempty(tline)
			C = textscan(tline,'%s','delimiter',',');
			C = C{1}';
			mets_dictionary = [mets_dictionary ;C];
			m = m + 1;
		end
	end
	fclose(fid);

    coculture.metNames = mets_dictionary(:  ,  end);

    %The last column is for human usage, not for program
    species_name_medium(end) = [];
    mets_dictionary(:  ,  end) = [];

    % coculture.mets is the model independent names of metabolites (used for the coculture)
    coculture.mets = mets_dictionary(:  ,  end);

%     % Klitgord data patch, see env_epis scripts
%     %  coculture.mets_alias is the one different than medium column := {a  A  c  ala_L};
%     %  coculture.mets_alias_hashid is the index of medium id(row number) := [1  1  3    5  ];
% % SEED MODELS does not have alias name problem
    coculture.mets_alias = {};
    coculture.mets_alias_hashid = [];
    % Preparing the hash table for alias names
    for i=1:length(coculture.mets)
        dic = mets_dictionary(i,:) ;
        %Replace NaN into the canonical name (the one used in medium) to
        %avoid ismember function error message
        pos = isnan(cellfun(@mean,dic));
        dic(pos)=coculture.mets(i);

        % e.g. [a a1 a2 A A A],[A] gets [1 1 1 0 0 0] the first 3 species has alias names
        species_has_alias_or_not = ~ismember(dic , coculture.mets(i) );

        %detect at least one alias name
        if(sum(species_has_alias_or_not)>0    )
            species_id = find(species_has_alias_or_not);
            alias_names = dic(species_id);
            alias_names = unique(alias_names);
            alias_hashid = i*ones( 1 , length(alias_names) );
            if(i == 76)
               disp('')
            end
            coculture.mets_alias = [coculture.mets_alias  alias_names];
            coculture.mets_alias_hashid = [coculture.mets_alias_hashid  alias_hashid];

        end
    end
    coculture.mets_alias = coculture.mets_alias';
    coculture.mets_alias_hashid =coculture.mets_alias_hashid';

    % Now, for each species k, establish EX_rxn -> mets_env info
    for k=1:length(coculture.species)
        secreted_mets = models(k).EXmets;
        medium_metsID=[];

        species_index = strmatch(species(k) , species_name_medium, 'exact');

        species_mets = mets_dictionary(:  ,   species_index  );
        nan_pos = isnan(cellfun(@mean, species_mets     ));
        species_mets(nan_pos) = {'no_transporter'};
    %     ALSO DO THE OTHER WAY AROUND ?
        for i=1:length(secreted_mets)
            % the index of this secreted_mets in the medium
            clear tmp;
            tmp = strmatch(secreted_mets{i} , species_mets, 'exact');

            if (isempty(tmp))
                if ~strcmp(secreted_mets{i},'biomass[c]')
                    MM = secreted_mets{i};
                    disp([ models(k).description 'secreting nonexisting metabolite in the medium:  ' MM]);
                    disp('check dictionary')
                    stop;
                end
            else
                %this is the index of met_universe for EXrxns of species k
                medium_metsID = [medium_metsID ; tmp];
            end
        end
        % e.g. EXmetsMEDIUMid maps glc in models(1) as the 121th mets in the medium
        models(k).EXmetsMEDIUMid = medium_metsID;
    end
	coculture.mets = strrep(coculture.mets , '(e)' , '[env]' ) ;
	coculture.mets = strrep(coculture.mets , '[e]' , '[env]' ) ;
	% % Klitgord data patch,
	coculture.mets_alias = strrep(coculture.mets_alias , '[e]' , '[env]' ) ;
	coculture.mets_alias = strrep(coculture.mets_alias , '(e)' , '[env]' ) ;
    % Disable default uptake rates that comes with the FBA models
    for k=1:length(coculture.species)
       tmp = find(models(k).isEX);
       models(k).lb(  tmp  ) = zeros( size(tmp)  );
    end
end
